-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = _G
local ipairs, pairs = ipairs, pairs
local string = string
local format = string.format
local strsplit = string.split
local table = table
local wipe = table.wipe
local tinsert = table.insert
local tremove = table.remove
local floor = math.floor

-------------------------------------------------------------------------------
-- Localized Blizzard API
-------------------------------------------------------------------------------
local UnitClass, UnitLevel, UnitName = UnitClass, UnitLevel, UnitName
local GetItemInfo, GetItemGem = GetItemInfo, GetItemGem
local GetInventoryItemLink =  GetInventoryItemLink

local MAX_NUM_SOCKETS = MAX_NUM_SOCKETS

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local NAME = "GearGauge"
local GearGauge = LibStub("AceAddon-3.0"):NewAddon(NAME, "AceEvent-3.0")

local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local QTC = LibStub:GetLibrary("LibQTipClick-1.1")
local LBC = LibStub("LibBabble-Class-3.0"):GetLookupTable()

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(NAME, "enUS", true, dev)

local DataObj
local tooltip

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local TBC_MAXLEV = 60
local WOTLK_MAXLEV = 71
local IS_ENCHANTER = false

local EQUIPMENT = {
	["Head"]		= { minlev = 30,	name = _G["INVTYPE_HEAD"],		mult = 0.8125 },
	["Neck"]		= { minlev = 30,	name = _G["INVTYPE_NECK"],		mult = 0.55 },
	["Shoulder"]		= { minlev = 20,	name = _G["INVTYPE_SHOULDER"],		mult = 0.75 },
	["Chest"]		= { minlev = 1,		name = _G["INVTYPE_CHEST"],		mult = 1 },
	["Waist"]		= { minlev = 1,		name = _G["INVTYPE_WAIST"],		mult = 0.5625 },
	["Legs"]		= { minlev = 1,		name = _G["INVTYPE_LEGS"],		mult = 0.875 },
	["Feet"]		= { minlev = 1,		name = _G["INVTYPE_FEET"],		mult = 0.6875 },
	["Wrist"]		= { minlev = 1,		name = _G["INVTYPE_WRIST"],		mult = 0.4375 },
	["Hands"]		= { minlev = 1,		name = _G["INVTYPE_HAND"],		mult = 0.625 },
	["Finger0"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"],		mult = 0.55 },
	["Finger1"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"],		mult = 0.55 },
	["Trinket0"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"],		mult = 0.75 },
	["Trinket1"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"],		mult = 0.75 },
	["Back"]		= { minlev = 1,		name = _G["INVTYPE_CLOAK"],		mult = 0.48 },
	["MainHand"]		= { minlev = 1,		name = _G["INVTYPE_WEAPONMAINHAND"],	mult = 0.50 },
	["SecondaryHand"]	= { minlev = 1,		name = _G["INVTYPE_WEAPONOFFHAND"],	mult = 0.50 },
	["Ranged"]		= { minlev = 1,		name = _G["INVTYPE_RANGED"],		mult = 0.35 },
}

local ORDERED_EQUIPMENT = {
	[1] = "Head",		[2] = "Neck",		[3] = "Shoulder",
	[4] = "Chest",		[5] = "Waist",		[6] = "Legs",
	[7] = "Feet",		[8] = "Wrist",		[9] = "Hands",
	[10] = "Finger0",	[11] = "Finger1",	[12] = "Trinket0",
	[13] = "Trinket1",	[14] = "Back",		[15] = "MainHand",
	[16] = "SecondaryHand",	[17] = "Ranged"
}

local TWO_HANDED_WEAPON = {
	[L["Fishing Poles"]] = true,
	[L["Polearms"]] = true,
	[L["Staves"]] = true,
	[L["Two-Handed Axes"]] = true,
	[L["Two-Handed Maces"]] = true,
	[L["Two-Handed Swords"]] = true,
}
local EQ_SUM = 0

-------------------------------------------------------------------------------
-- Configuration options (future)
-------------------------------------------------------------------------------
local MAX_LEVEL = 80
local MIN_LEVEL = 1

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local char_equip = {}		-- Character's equipment. Generated when AddOn is loaded or state is changed.
local target_equip = {}		-- Target's equipment. Generated when AddOn is loaded or state is changed.

-------------------------------------------------------------------------------
-- Table cache
-------------------------------------------------------------------------------
local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable()
		local tbl = tremove(table_cache) or {}
		return tbl
	end

	-- Cleans the table and stores it in the cache
	function ReleaseTable(tbl)
		if not tbl then return end
		wipe(tbl)
		tinsert(table_cache, tbl)
	end
end	-- do

-------------------------------------------------------------------------------
-- Local functions
-------------------------------------------------------------------------------
local function AverageItemLevels(unit)
	unit = unit or "player"

	local unit_level = UnitLevel(unit)
	local level_cnt = 0
	local item_cnt = 0


	for slot in pairs(EQUIPMENT) do
		if unit == "player" then
			local item = char_equip[slot]

			if item.level > 0 then
				level_cnt = level_cnt + item.level
				item_cnt = item_cnt + 1
			end
		else
			local link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)

			if link then
				local _, _, _, item_level, _, _, _, _, _, _ = GetItemInfo(link)

				if item_level > 0 then
					level_cnt = level_cnt + item_level
					item_cnt = item_cnt + 1
				end
			end
		end
	end
	if item_cnt == 0 then return 0 end

	local avg = floor(level_cnt / item_cnt)
	return avg
end

-------------------------------------------------------------------------------
-- Item status generation
-------------------------------------------------------------------------------
local BuildGearData
do
	local prev_stype	-- Placeholder in case previous slot contained a two-handed weapon.

	-------------------------------------------------------------------------------
	-- Check to see if an empty slot is OK given the unit's level/class.
	-------------------------------------------------------------------------------
	local function IsEmptySlotOK(slot, level, class)
		if slot == "SecondaryHand"  then
			if level < 5 and (class == LBC["Paladin"] or class == LBC["Shaman"] or class == LBC["Warrior"]) then	-- Shields
				return true
			else
				return false
			end
		end

		if slot == "Ranged" then
			if  level < 52 and (class == LBC["Druid"] or class == LBC["Paladin"] or class == LBC["Shaman"]) then	-- Idols, Librams, Totems
				return true
			elseif level < 5 and (class == LBC["Mage"] or class == LBC["Priest"] or class == LBC["Warlock"]) then	-- Wands
				return true
			else
				return false
			end
		end
		return EQUIPMENT[slot].minlev > level
	end


	-------------------------------------------------------------------------------
	-- Tooltip "scanning" for empty gem slots. Initial code by Tristanian.
	-------------------------------------------------------------------------------
	local GemTooltip = CreateFrame("GameTooltip", "GearGaugeGemTooltip", UIParent, "GameTooltipTemplate")
	GemTooltip:Hide()

	local function GetEmptySockets(item)
		if not item then return end

		if not item.link then
			item.empty_meta, item.empty_red, item.empty_yellow, item.empty_blue = 0, 0, 0, 0
			return
		end
		local i, emptyMetaSockets, emptyRedSockets, emptyBlueSockets, emptyYellowSockets = nil, 0, 0, 0, 0
		local emptyMetaTexture = "Interface\\ItemSocketingFrame\\UI-EmptySocket-Meta"
		local emptyRedTexture = "Interface\\ItemSocketingFrame\\UI-EmptySocket-Red"
		local emptyBlueTexture = "Interface\\ItemSocketingFrame\\UI-EmptySocket-Blue"
		local emptyYellowTexture = "Interface\\ItemSocketingFrame\\UI-EmptySocket-Yellow"

		-- clear the textures first
		for i = 1, MAX_NUM_SOCKETS do
			if _G["GearGaugeGemTooltipTexture"..i] then
				_G["GearGaugeGemTooltipTexture"..i]:SetTexture("")
			end
		end
		GemTooltip:SetOwner(UIParent, "ANCHOR_NONE")
		GemTooltip:ClearLines()
		GemTooltip:SetHyperlink(item.link)

		for i = 1, MAX_NUM_SOCKETS do
			local temp = _G["GearGaugeGemTooltipTexture"..i]:GetTexture()

			if temp and temp == emptyMetaTexture then emptyMetaSockets = emptyMetaSockets + 1 end
			if temp and temp == emptyRedTexture then emptyRedSockets = emptyRedSockets + 1 end
			if temp and temp == emptyYellowTexture then emptyYellowSockets = emptyYellowSockets + 1 end     
			if temp and temp == emptyBlueTexture then emptyBlueSockets = emptyBlueSockets + 1 end     
		end
		item.empty_meta, item.empty_red, item.empty_yellow, item.empty_blue = emptyMetaSockets, emptyRedSockets, emptyYellowSockets, emptyBlueSockets
	end

	-- Item subtypes which are not enchantable
	local NON_ENCHANTABLE = {
		[L["Idols"]] = true,
		[L["Librams"]] = true,
		[L["Miscellaneous"]] = true,
		[L["Other"]] = true,
		[L["Sigils"]] = true,
		[L["Thrown"]] = true,
		[L["Totems"]] = true,
		[L["Wands"]] = true,
	}

	-------------------------------------------------------------------------------
	-- Assign values to an item for later use.
	-------------------------------------------------------------------------------
	local function ConstructItem(item_link, slot, unit)
		local unit_level = UnitLevel(unit)
		local item = AcquireTable()

		item.slot, item.rating, item.level = slot, 0, 0

		if not item_link then
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]
			local is_ok = IsEmptySlotOK(slot, unit_level, UnitClass(unit))

			if not two_handed and not is_ok then
				item.rating = - EQUIPMENT[slot].mult
			end
			item.empty_ok = is_ok
			prev_stype = nil
			-- Get gem information
			GetEmptySockets(item)
			return item
		end
		item.name, item.link, item.rarity, item.level, item.minlev, item.type, item.stype, item.stacksize, item.equiploc, item.texture = GetItemInfo(item_link)
		prev_stype = item.stype

		-- Get gem information
		GetEmptySockets(item)

		-- Set the level of Heirloom items depending on the player's level
		if item.rarity == 7 then
			if unit_level <= TBC_MAXLEV then
				item.level = unit_level
			elseif unit_level > TBC_MAXLEV and unit_level <= WOTLK_MAXLEV then
				item.level = 3 * (unit_level - 64) + 90
			elseif unit_level > WOTLK_MAXLEV then
				item.level = 4 * (unit_level - 73) + 142
			end
		end
		if item.level <= 1 then return item end

		local points = ((item.level / unit_level) + ((item.rarity - 1) * 0.5)) * EQUIPMENT[slot].mult
		local rating = points
		local _, _, ench_id = strsplit(":", item_link)		-- Grab the item's enchant id, if enchanted.

		if slot == "Waist" then
			item.ench_id = -1
		else
			if ench_id == nil or ench_id == "0" then
				if ((slot == "Finger0" or slot == "Finger1") and IS_ENCHANTER) or not NON_ENCHANTABLE[item.stype] then
					rating = rating - (points * 0.25)
				else
					item.ench_id = -1
				end
			else
				item.ench_id = ench_id
				rating = rating + (points * 0.25)
			end
		end

		local empty_sockets = item.empty_meta + item.empty_red + item.empty_yellow + item.empty_blue

		for i = 1, empty_sockets do
			local name, gem = GetItemGem(item_link, i)
			item["gem"..i] = gem

			if not gem then
				rating = rating - (points * 0.125)
			else
				rating = rating + (points * 0.125)
			end
		end
		item.rating = rating

		if TWO_HANDED_WEAPON[item.stype] then
			item.rating = item.rating * 2
		end
		return item
	end

	-------------------------------------------------------------------------------
	-- List of items equipped by unit.
	-------------------------------------------------------------------------------
	function BuildGearData(unit, item_list)
		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			ReleaseTable(item_list[slot])
			wipe(item_list)
		end
		local rating = 0

		prev_stype = nil

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)
			local item = ConstructItem(link, slot, unit)

			rating = rating + item.rating
			item_list[slot] = item
		end
		return rating
	end
end -- do

do
	-- Table used to gather information about a target, to be used elsewhere.
	local target = {}

	function GearGauge:Scan(unit)
		if InCombatLockdown() then return end
		if not (UnitExists(unit) and UnitIsPlayer(unit) and UnitLevel(unit) >= MIN_LEVEL and UnitLevel(unit) <= MAX_LEVEL) then return end

		local unit_name = UnitName(unit)
		NotifyInspect(unit)
		wipe(target)

		-------------------------------------------------------------------------------
		-- For future implementation
		-------------------------------------------------------------------------------
		--		target.name = UnitName(unit)
		--		target.level = UnitLevel(unit)
		--		target.class = UnitClass(unit)
		--		target.gender = UnitSex(unit)
		--		target.race = UnitRace(unit)
		--		target.guild = GetGuildInfo(unit) or "("..NOT_APPLICABLE..")"
		--		target.faction = UnitFactionGroup(unit)
		--		target.updated = time()

		target.item_list = AcquireTable()
		target.rating = BuildGearData(unit, target.item_list)

		-- Wipe the target's equipment list
		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			ReleaseTable(target.item_list[slot])
		end
		ReleaseTable(target.item_list)

		local scan_floor = -EQ_SUM * UnitLevel(unit)

		if target.rating <= scan_floor then
			GearGauge:Print(target.name..": "..EMPTY)
			return
		end
		local equip_level = AverageItemLevels(unit)

		if unit_name ~= UnitName("player") then
			GearGauge:Print(unit_name.." "..RATING..": "..format("%.2f", target.rating)..". "..ITEMS.." "..LEVEL..": "..equip_level)
		end
	end
end -- do

function GearGauge:Print(msg)
	if not msg then return end
	DEFAULT_CHAT_FRAME:AddMessage("|cffffff00"..NAME.."|r: "..msg)
end

-------------------------------------------------------------------------------
-- Event handlers
-------------------------------------------------------------------------------
local function UpdateData()
	DataObj.text = format("%.2f", BuildGearData("player", char_equip))
end

function GearGauge:PLAYER_LEVEL_UP()
	UpdateData()
end

function GearGauge:PLAYER_TARGET_CHANGED()
	if InCombatLockdown() then return end

	if UnitExists("target") and UnitIsPlayer("target") and CheckInteractDistance("target", 1) then
		self:Scan("target")
	end
end

function GearGauge:UNIT_INVENTORY_CHANGED()
	UpdateData()
end

-------------------------------------------------------------------------------
-- Tooltip functions
-------------------------------------------------------------------------------
local GetItemStatus
do
	-- Returns a string based on whether or not the item is enchanted
	function GetItemStatus(item)
		if not item.link then return HELPFRAME_ITEM_TITLE end

		local ench_id = item.ench_id
		local retstr, comma = "", ""

		if ench_id ~= -1 and (ench_id == nil or ench_id == "0") then
			retstr = retstr..ENSCRIBE
			comma = KEY_COMMA.." "
		end

		if (item.empty_meta + item.empty_red + item.empty_yellow + item.empty_blue) > 0 then
			retstr = retstr..comma..SOCKET_GEMS
			comma = KEY_COMMA.." "
		end
		return retstr
	end
end --do

local function GetTipAnchor(frame)
	local x, y = frame:GetCenter()

	if not x or not y then return "TOPLEFT", "BOTTOMLEFT" end

	local hhalf = (x > UIParent:GetWidth()*2/3) and "RIGHT" or (x < UIParent:GetWidth()/3) and "LEFT" or ""
	local vhalf = (y > UIParent:GetHeight()/2) and "TOP" or "BOTTOM"
	return vhalf..hhalf, frame, (vhalf == "TOP" and "BOTTOM" or "TOP")..hhalf
end

local function ItemOnEnter(event, cell, arg)
	QTC.OnEnter(event, cell, arg)

	GameTooltip:SetOwner(tooltip, "ANCHOR_NONE")
	GameTooltip:ClearAllPoints()
	GameTooltip:SetPoint(GetTipAnchor(tooltip))
	GameTooltip:ClearLines()
	GameTooltip:SetHyperlink(arg)
	GameTooltip:Show()
end

local function ItemOnLeave(event, cell, arg)
	QTC.OnLeave(event, cell, arg)
	GameTooltip:Hide()
end

local function BuildTooltip(anchor, unit)
	unit = unit or "player"

	tooltip = QTC:Acquire(NAME.."Tooltip", 3, "LEFT", "LEFT", "LEFT")
	tooltip:SetCallback("OnEnter", ItemOnEnter)
	tooltip:SetCallback("OnLeave", ItemOnLeave)

	tooltip:Clear()
	tooltip:SetAutoHideDelay(0.25, anchor)
	tooltip:SmartAnchorTo(anchor)

	if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(tooltip, true) end

	local rating_tot = 0
	local equip_level = AverageItemLevels(unit)
	local line, column

	line, column = tooltip:AddNormalHeader()
	tooltip:SetNormalCell(line, column, LEVEL..": "..UnitLevel(unit)..KEY_COMMA.." "..ITEMS.." "..LEVEL..": "..equip_level, "CENTER", 3)
	tooltip:AddSeparator()
	line, column = tooltip:AddNormalHeader("", RATING, ADDON_MISSING)

	tooltip:SetNormalCell(line, 1, ITEMS, "CENTER")
	tooltip:AddSeparator()

	local two_handed = nil
	local prev_stype = nil

	for idx, slot in ipairs(ORDERED_EQUIPMENT) do
		local item = char_equip[slot]

		rating_tot = rating_tot + item.rating

		if item.link then
			local r, g, b, hex = GetItemQualityColor(item.rarity)

			line, column = tooltip:AddLine()
			tooltip:SetLineColor(line, r, g, b)
			tooltip:SetCell(line, 1, "|T"..item.texture..":25:25|t"..item.name, item.link)
			tooltip:SetNormalCell(line, 2, format("%.2f", item.rating))
			tooltip:SetNormalCell(line, 3, GetItemStatus(item))
			prev_stype = item.stype
		else
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]

			if not two_handed  then
				line, column = tooltip:AddLine()

				if item.empty_ok then
					tooltip:SetLineColor(line)
				else
					tooltip:SetLineColor(line, 1)
				end
				tooltip:SetNormalCell(line, 1, EQUIPMENT[slot].name, nil, nil, 1, nil, 20)
				tooltip:SetNormalCell(line, 2, format("%.2f", item.rating))
				tooltip:SetNormalCell(line, 3, GetItemStatus(item))
				prev_stype = nil
			end
		end
	end
	tooltip:AddSeparator()

	line, column = tooltip:AddLine()

	tooltip:SetNormalCell(line, 2, format("%.2f", rating_tot))
	tooltip:Show()
end

local function ShowTooltip(self)
	BuildTooltip(self)
end

-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
function GearGauge:OnInitialize()
	-- Set the numeric ID and texture strings for the equipment slots
	for slot in pairs(EQUIPMENT) do
		EQUIPMENT[slot].id, EQUIPMENT[slot].texture = GetInventorySlotInfo(slot.."Slot")
		EQ_SUM = EQ_SUM + EQUIPMENT[slot].mult
	end

	-- Determine if the toon is an enchanter
	local enchanting_sklname = GetSpellInfo(7411)
	for idx = 1, GetNumSkillLines() do
		local skill = GetSkillLineInfo(idx)
		if skill == enchanting_sklname then
			IS_ENCHANTER = true
			break
		end
	end

	DataObj = LDB:NewDataObject("GearGauge", {
		type	= "data source",
		label	= "GearGauge",
		text	= "0",
		icon	= "Interface\\Icons\\INV_Shield_36",
		OnEnter = function(self) ShowTooltip(self) end,
		OnLeave = function() end,
	})
end

function GearGauge:OnEnable()
	UpdateData()
	GearGauge:RegisterEvent("PLAYER_LEVEL_UP")
	GearGauge:RegisterEvent("PLAYER_TARGET_CHANGED")
	GearGauge:RegisterEvent("UNIT_INVENTORY_CHANGED")
end

function GearGauge:OnDisable()
	GearGauge:UnregisterEvent("PLAYER_LEVEL_UP")
	GearGauge:UnregisterEvent("PLAYER_TARGET_CHANGED")
	GearGauge:UnregisterEvent("UNIT_INVENTORY_CHANGED")
end
